__author__ = 'lifu'

import theano
import argparse
import numpy as np
import pickle
import pdb
from example import example
from model import model
from features import features
import lasagne
import sys
import random
import cPickle
import utils
import nltk


def add_features(example, args, features):
    
    example.features.extend(features.compute_features(example))



def get_data_one(s, target, args, features, wordsi, params=None):
 
    s = " ".join(nltk.word_tokenize(s)).lower()
    target = " ".join(nltk.word_tokenize(target)).lower()
   
    e = example(s, target, 0., 0., 0., words=words, params=params)
    add_features(e, args, features)

    return [e]



def get_data(fname, args, features, words, params=None):
    f = open(fname,'r')
    lines = f.readlines()
    f.close()
    
    examples = []
    for i in lines:
          i = i.split('\t')
       
          sents = nltk.word_tokenize(i[0])
          sents = " ".join(sents).lower()
          t = i[1]
          t = nltk.word_tokenize(t)
          t = " ".join(t).lower()

          e = example(sents, t, 0, 0, 0, words=words, params=params)
          add_features(e, args, features)
    
          examples.append(e)

    return examples, lines


def get_data_test(fname, args, features, words, params=None):
    f = open(fname,'r')
    lines = f.readlines()
    f.close()

    O = []
    R = []
    I = []
    examples = []
    for i in lines:
          i = i.split('\t')

          s = i[0]
          target = i[1]
          s = " ".join(nltk.word_tokenize(s)).lower()
          target = " ".join(nltk.word_tokenize(target)).lower()

          e = example(s, target, 0, 0, 0, words=words, params=params)
          add_features(e, args, features)

          examples.append(e)
          O.append(float(i[5])/10)
          R.append(float(i[6])/10)
          I.append(float(i[7])/10)

    return examples, lines, O, R, I


    
def str2act(v):
    if v is None:
        return v
    if v.lower() == "tanh":
        return lasagne.nonlinearities.tanh
    if v.lower() == "linear":
        return lasagne.nonlinearities.linear
    if v.lower() == "relu":
        return lasagne.nonlinearities.rectify
    if v.lower() == "sigmoid":
        return lasagne.nonlinearities.sigmoid
    raise ValueError('A type that was supposed to be a learner is not.')

np.set_printoptions(linewidth=160)
random.seed(1)
np.random.seed(1)

parser = argparse.ArgumentParser(description='story evaluation')
parser.add_argument('-nntype', help='lstm,avg,gan,bilistm,bigan,lstmavg', default='avg')
parser.add_argument('-LW', type=float, default=0.)
parser.add_argument('-LC', type=float, default=0.)
parser.add_argument('--add_features', action='store_true' )
parser.add_argument('-epochs', type=int, default=10)
parser.add_argument('-batchsize', type=int, default=100)
parser.add_argument('-model', type=int, default=1)
parser.add_argument('-act', default='relu')
parser.add_argument('-outfile', default="params")
parser.add_argument('--save', action='store_true')
parser.add_argument('--evaluate', action='store_false')
parser.add_argument('-single', type=int, default=-1)
parser.add_argument('-ngrams', type=int, default=2)
parser.add_argument('-min_count', type=int)
parser.add_argument('-dim', type=int, default=300)
parser.add_argument('--sumlayer', action='store_true')
parser.add_argument('--random_init', action='store_true')



parser.add_argument('--keybordinput', action='store_true')
parser.add_argument('--file')
parser.add_argument('--resultfile')

args = parser.parse_args()
args.act = str2act(args.act)

print " ".join(sys.argv)

features = features()


args.nfeatures = 4


words, We = cPickle.load(open('ngram_dic.pkl', 'rb'))


model = model(We, args)

p_values = cPickle.load(open('ilcontext.pkl', 'rb'))
lasagne.layers.set_all_param_values(model.layers, p_values)

"""
############### This part is for evaluating the performance on the dev set in the paper
examples, lines, O, R, I = get_data_test('val.kgdetok', args, features, words, params = args)



#evaluate the stories
pred_o, pred_r, pred_i = model.evaluate(examples, args)

from scipy.stats import spearmanr
from scipy.stats import pearsonr
#print pred_o
#print O

print "spearman: ", spearmanr(O, pred_o)[0], spearmanr(R, pred_r)[0], spearmanr(I, pred_i)[0]
print "pearson: ", pearsonr(O, pred_o)[0], pearsonr(R, pred_r)[0], pearsonr(I, pred_i)[0]

"""
if args.keybordinput:
      try: 
           while(1):
                  s = raw_input("Please input the story context:\n")
                  print ""
                  target = raw_input("Please input the story ending:\n")
                  examples = get_data_one(s, target, args, features, words, params=None)
                  pred_o, pred_r, pred_i = model.evaluate(examples, args)
                  print ""
                  print "Overall score:", 10*pred_o[0],  "\tRelevance score:", 10*pred_r[0], "\tInterestingness score:", 10*pred_i[0]
                  print ""

      except KeyboardInterrupt:
           print "finishing the demo" 

else:
      examples, lines = get_data(args.file, args, features, words, params = args)
      pred_o, pred_r, pred_i = model.evaluate(examples, args)
      f = open(args.resultfile, 'w')
      for i, line in enumerate(lines):   
                f.write(line.strip() + '\t'+ str(10*pred_o[i]) + '\t'+ str(10*pred_r[i]) + '\t'+ str(10*pred_i[i]) + '\n')
      f.close()
