import theano
import numpy as np
from theano import tensor as T
from theano import config
import pdb
import time
import utils
from lstm_average_layer import lstm_average_layer
from lasagne_sum_layer import lasagne_sum_layer
from lasagne_average_layer import lasagne_average_layer
from lasagne_bilinear_layer import lasagne_bilinear_layer
import lasagne
import sys
import cPickle
from scipy.stats import spearmanr
from scipy.stats import pearsonr

class model(object):
    # takes list of seqs, puts them in a matrix
    # returns matrix of seqs and mask
    def prepare_data(self, list_of_seqs):
        lengths = [len(s) for s in list_of_seqs]
        n_samples = len(list_of_seqs)
        maxlen = np.max(lengths)
        x = np.zeros((n_samples, maxlen)).astype('int32')
        x_mask = np.zeros((n_samples, maxlen)).astype(theano.config.floatX)
        for idx, s in enumerate(list_of_seqs):
            x[idx, :lengths[idx]] = s
            x_mask[idx, :lengths[idx]] = 1.
        x_mask = np.asarray(x_mask, dtype=config.floatX)
        return x, x_mask

    def prepare_batch(self, batch, params):

        features = None

        if len(batch[0].features) > 0:
            features = np.zeros((len(batch), len(batch[0].features) + 1))

        g1 = []
        g2 = []
        for i in batch:
            g1.append(i.embeddings_context)
            g2.append(i.embeddings_target)

        if type(features) is np.ndarray:
            for i in range(len(batch)):
                for j in range(len(batch[i].features)):
                    #pdb.set_trace()
                    features[i][j] = batch[i].features[j]
                features[i][j+1] = 1
            features = np.asarray(features, dtype=config.floatX)

        g1x, g1mask = self.prepare_data(g1)
        g2x, g2mask = self.prepare_data(g2)

        return g1x, g1mask, g2x, g2mask, features

    def saveParams(self, fname):
        f = file(fname, 'wb')
        values = lasagne.layers.get_all_param_values(self.layers)
        # values = [i.get_value() for i in values]
        cPickle.dump(values, f, protocol=cPickle.HIGHEST_PROTOCOL)
        f.close()

    def get_minibatches_idx(self, n, minibatch_size, shuffle=False):
        idx_list = np.arange(n, dtype="int32")

        if shuffle:
            np.random.shuffle(idx_list)

        minibatches = []
        minibatch_start = 0
        for i in range(n // minibatch_size):
            minibatches.append(idx_list[minibatch_start:
            minibatch_start + minibatch_size])
            minibatch_start += minibatch_size

        if (minibatch_start != n):
            # Make a minibatch out of what is left
            minibatches.append(idx_list[minibatch_start:])

        return zip(range(len(minibatches)), minibatches)


    def __init__(self, We_initial, params):

        initial_We = theano.shared(np.asarray(We_initial, dtype=config.floatX))
        We = theano.shared(np.asarray(We_initial, dtype=config.floatX))

        # symbolic params
        g1idx = T.imatrix()
        g1mask = T.matrix()
        g2idx = T.imatrix()
        g2mask = T.matrix()
        scores = T.matrix()

        if params.single >= 0:
            scores = T.vector()

        features = T.matrix()

        # get embeddings
        l_in = lasagne.layers.InputLayer((None, None))
        l_mask = lasagne.layers.InputLayer(shape=(None, None))
        l_emb = lasagne.layers.EmbeddingLayer(l_in, input_size=We.get_value().shape[0],
                                                  output_size=We.get_value().shape[1], W=We)

        if params.nntype == "lstm":
            l_lstm = lasagne.layers.LSTMLayer(l_emb, 300, peepholes=True, learn_init=False,
                                                  mask_input=l_mask)

            l_out = lasagne.layers.SliceLayer(l_lstm, -1, 1)

        elif params.nntype == "lstmavg":
            l_lstm = lasagne.layers.LSTMLayer(l_emb, 300, peepholes=True, learn_init=False,
                                                  mask_input=l_mask)

            l_out = lasagne_average_layer([l_lstm, l_mask], tosum=params.sum)

        elif params.nntype == "gan":
            l_lstm = lstm_average_layer(l_emb, 300, peepholes=True, learn_init=False,
                                            mask_input=l_mask, model_type=1)

            l_out = lasagne_average_layer([l_lstm, l_mask], tosum=False)

        elif params.nntype == "bigan":
            l_lstm = lstm_average_layer(l_emb, 300, peepholes=True, learn_init=False,
                                            mask_input=l_mask, model_type=params.model_type)
            l_lstmb = lstm_average_layer(l_emb, 300, peepholes=True, learn_init=False,
                                             mask_input=l_mask, backwards=True)

            if not params.sumlayer:
                l_concat = lasagne.layers.ConcatLayer([l_lstm, l_lstmb], axis=2)
                l_out = lasagne.layers.DenseLayer(l_concat, 300, num_leading_axes=-1,
                                                  nonlinearity=lasagne.nonlinearities.tanh)
                l_out = lasagne_average_layer([l_out, l_mask], tosum=False)
            else:
                l_out = lasagne_sum_layer([l_lstm, l_lstmb])
                l_out = lasagne_average_layer([l_out, l_mask], tosum=False)

        elif params.nntype == "bilstmavg":
            l_rnn = lasagne.layers.LSTMLayer(l_emb, 300, learn_init=False,
                                                 mask_input=l_mask)
            l_rnnb = lasagne.layers.LSTMLayer(l_emb, 300, learn_init=False,
                                                  mask_input=l_mask, backwards=True)

            if not params.sumlayer:
                l_concat = lasagne.layers.ConcatLayer([l_rnn, l_rnnb], axis=2)
                l_out = lasagne.layers.DenseLayer(l_concat, 300, num_leading_axes=-1,
                                                  nonlinearity=lasagne.nonlinearities.tanh)
                l_out = lasagne_average_layer([l_out, l_mask], tosum=params.sum)
            else:
                l_out = lasagne_sum_layer([l_rnn, l_rnnb])
                l_out = lasagne_average_layer([l_out, l_mask], tosum=params.sum)

        elif params.nntype == "bilstm":
            l_rnn = lasagne.layers.LSTMLayer(l_emb, 300, learn_init=False,
                                                 mask_input=l_mask)
            l_rnnb = lasagne.layers.LSTMLayer(l_emb, 300, learn_init=False,
                                                  mask_input=l_mask, backwards=True)

            if not params.sumlayer:
                l_outf = lasagne.layers.SliceLayer(l_rnn, -1, 1)
                l_outb = lasagne.layers.SliceLayer(l_rnnb, -1, 1)

                l_concat = lasagne.layers.ConcatLayer([l_outf, l_outb], axis=1)
                l_out = lasagne.layers.DenseLayer(l_concat, 300, nonlinearity=lasagne.nonlinearities.tanh)
            else:
                l_out = lasagne_sum_layer([l_rnn, l_rnnb])
                l_out = lasagne.layers.DenseLayer(l_out, 300, nonlinearity=lasagne.nonlinearities.tanh)

        elif params.nntype == "avg":
            l_out = lasagne_average_layer([l_emb, l_mask], tosum=False)

        em1 = lasagne.layers.get_output(l_out, {l_in: g1idx, l_mask: g1mask}, deterministic=True)
        em2 = lasagne.layers.get_output(l_out, {l_in: g2idx, l_mask: g2mask}, deterministic=True)

        g1_dot_g2 = em1 * em2
        g1_abs_g2 = abs(em1 - em2)

        if params.model == 1:
            l_in_1 = lasagne.layers.InputLayer((None, 300))
            l_in_2 = lasagne.layers.InputLayer((None, 300))
            l_in_3 = lasagne.layers.InputLayer((None, 300))
            l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3])
            if params.nfeatures > 0:
                l_in_4 = lasagne.layers.InputLayer((None, params.nfeatures))
                l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3, l_in_4])
            l_dense1 = lasagne.layers.DenseLayer(l_sum, 300, nonlinearity=params.act)
            if params.single >= 0:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 1, nonlinearity=lasagne.nonlinearities.linear)
            else:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 3, nonlinearity=lasagne.nonlinearities.linear)
            if params.nfeatures > 0:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em2, l_in_2: g1_dot_g2, l_in_3: g1_abs_g2, l_in_4: features})
            else:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em2, l_in_2: g1_dot_g2, l_in_3: g1_abs_g2})
        elif params.model == 2:
            l_in_1 = lasagne.layers.InputLayer((None, 300))
            l_in_2 = lasagne.layers.InputLayer((None, 300))
            l_in_3 = lasagne.layers.InputLayer((None, 300))
            l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3])
            if params.nfeatures > 0:
                l_in_4 = lasagne.layers.InputLayer((None, params.nfeatures))
                l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3, l_in_4])
            l_dense1 = lasagne.layers.DenseLayer(l_sum, 300, nonlinearity=params.act)
            if params.single >= 0:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 1, nonlinearity=lasagne.nonlinearities.linear)
            else:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 3, nonlinearity=lasagne.nonlinearities.linear)
            if params.nfeatures > 0:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em1, l_in_2: g1_dot_g2, l_in_3: g1_abs_g2, l_in_4: features})
            else:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em1, l_in_2: g1_dot_g2, l_in_3: g1_abs_g2})
        elif params.model == 3:
            l_in_1 = lasagne.layers.InputLayer((None, 300))
            l_in_2 = lasagne.layers.InputLayer((None, 300))
            l_in_3 = lasagne.layers.InputLayer((None, 300))
            l_in_4 = lasagne.layers.InputLayer((None, 300))
            l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3, l_in_4])
            if params.nfeatures > 0:
                l_in_5 = lasagne.layers.InputLayer((None, params.nfeatures))
                l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3, l_in_4, l_in_5])
            l_dense1 = lasagne.layers.DenseLayer(l_sum, 300, nonlinearity=params.act)
            if params.single >= 0:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 1, nonlinearity=lasagne.nonlinearities.linear)
            else:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 3, nonlinearity=lasagne.nonlinearities.linear)
            if params.nfeatures > 0:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em1, l_in_2: em2, l_in_3: g1_dot_g2, l_in_4: g1_abs_g2, l_in_5: features})
            else:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em1, l_in_2: em2, l_in_3: g1_dot_g2, l_in_4: g1_abs_g2})
        elif params.model == 4:
            l_in_1 = lasagne.layers.InputLayer((None, 300))
            l_in_2 = lasagne.layers.InputLayer((None, 300))
            l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2])
            if params.nfeatures > 0:
                l_in_3 = lasagne.layers.InputLayer((None, params.nfeatures))
                l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2, l_in_3])
            l_dense1 = lasagne.layers.DenseLayer(l_sum, 300, nonlinearity=params.act)
            if params.single >= 0:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 1, nonlinearity=lasagne.nonlinearities.linear)
            else:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 3, nonlinearity=lasagne.nonlinearities.linear)
            if params.nfeatures > 0:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: g1_dot_g2, l_in_2: g1_abs_g2, l_in_3:features})
            else:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: g1_dot_g2, l_in_2: g1_abs_g2})
        elif params.model == 5: #just logistic regression
            l_in = lasagne.layers.InputLayer((None, params.nfeatures))
            if params.single >= 0:
                l_out2 = lasagne.layers.DenseLayer(l_in, 1, nonlinearity=lasagne.nonlinearities.linear)
            else:
                l_out2 = lasagne.layers.DenseLayer(l_in, 3, nonlinearity=lasagne.nonlinearities.linear)
            prediction = lasagne.layers.get_output(l_out2, {l_in: features})
        elif params.model == 6:
            l_in_1 = lasagne.layers.InputLayer((None, 300))
            l_sum = l_in_1
            if params.nfeatures > 0:
                l_in_2 = lasagne.layers.InputLayer((None, params.nfeatures))
                l_sum = lasagne.layers.ConcatLayer([l_in_1, l_in_2])
            l_dense1 = lasagne.layers.DenseLayer(l_sum, 300, nonlinearity=params.act)
            if params.single >= 0:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 1, nonlinearity=lasagne.nonlinearities.linear)
            else:
                l_out2 = lasagne.layers.DenseLayer(l_dense1, 3, nonlinearity=lasagne.nonlinearities.linear)
            if params.nfeatures > 0:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em2, l_in_2: features})
            else:
                prediction = lasagne.layers.get_output(l_out2, {l_in_1: em2})

        cost = abs(prediction-scores)
        cost = cost.sum(axis=1)
        network_params = lasagne.layers.get_all_params([l_out, l_out2], trainable=True)
        network_params.pop(0)
        self.all_params = lasagne.layers.get_all_params([l_out, l_out2], trainable=True)
        #print self.all_params

        self.layers = [l_out, l_out2]

        # regularization
        #l2 = 0.5 * params.LC * sum(lasagne.regularization.l2(x) for x in network_params)
        word_reg = 0.5 * params.LW * lasagne.regularization.l2(We - initial_We)
        #cost = T.mean(cost) + l2 + word_reg
        cost = T.mean(cost) + word_reg
        self.scoring_function = theano.function([g1idx, g1mask, g2idx, g2mask, features], prediction, on_unused_input='warn')

        # updates
        grads = theano.gradient.grad(cost, self.all_params)
        updates = lasagne.updates.adam(grads, self.all_params, 0.001)

        self.train_function = theano.function([g1idx, g1mask, g2idx, g2mask, scores, features], cost, updates=updates, on_unused_input='warn')
        cost = T.mean(cost)

        #print "Num Params:", utils.countParameters(self)

    def mean_absolute_error(self, scores1, scores2):
        vals = []
        for i in range(len(scores1)):
            vals.append(abs(scores1[i]-scores2[i]))
        return np.mean(vals)

    def evaluate(self, data, params):

        kf = self.get_minibatches_idx(len(data), params.batchsize, shuffle=False)
     
        pred_o = []; pred_r = []; pred_i = []
        for _, index in kf:
            batch = [data[t] for t in index]
            g1x, g1mask, g2x, g2mask, features = self.prepare_batch(batch, params)
            preds = self.scoring_function(g1x, g1mask, g2x, g2mask, features)
            for i in range(preds.shape[0]):
                pred_o.append(preds[i, :][0])
                pred_r.append(preds[i, :][1])
                pred_i.append(preds[i, :][2])


        return pred_o, pred_r, pred_i 
